#!/usr/bin/env python3
#
# Copyright 2017-2018 Ettus Research, a National Instruments Company
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
RX samples to file using Python API
"""

import argparse
import numpy as np
import uhd


def parse_args():
    """Parse the command line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--args", default="", type=str)
    parser.add_argument("-o", "--output-file", type=str, required=True)
    parser.add_argument("-f", "--freq", type=float, required=True)
    parser.add_argument("-r", "--rate", default=1e6, type=float)
    parser.add_argument("-d", "--duration", default=5.0, type=float)
    parser.add_argument("-c", "--channels", default=0, nargs="+", type=int)
    parser.add_argument("-g", "--gain", type=int, default=10)
    parser.add_argument("-n", "--numpy", default=False, action="store_true",
                        help="Save output file in NumPy format (default: No)")
    return parser.parse_args()


def main():
    """RX samples and write to file"""
    args = parse_args()
    usrp = uhd.usrp.MultiUSRP(args.args)
    num_samps = int(np.ceil(args.duration*args.rate))
    if not isinstance(args.channels, list):
        args.channels = [args.channels]
    samps = usrp.recv_num_samps(num_samps, args.freq, args.rate, args.channels, args.gain)
    with open(args.output_file, 'wb') as out_file:
        if args.numpy:
            np.save(out_file, samps, allow_pickle=False, fix_imports=False)
        else:
            samps.tofile(out_file)

if __name__ == "__main__":
    main()



#!D:/bld/uhd_1600087591633/_h_env/python.exe
#
# Copyright 2018 Ettus Research, a National Instruments Company
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Download image files required for USRPs.

Usage: The `uhd_images_downloader` should work, "out of the box", with no command line arguments.
Assuming your computer has an internet connection to [files.ettus.com], simply run the utility
every time you update UHD, and the installed images for your devices should always be up to date.


Images will be downloaded on a per-target basis. That is, there are image packages for a desired
device and configuration. Users can specify which image packages they would plan to use. To see a
list of available targets, run `uhd_images_downloader --list-targets`. The left column of the
printout will be a list of available image archives. From there, you can construct a regular
expression which matches to the targets you wish to download. For example, in order to download all
image packages related to the X300 product line, users may run
`uhd_images_downloader --types x3.*`.


The `uhd_images_downloader` uses a manifest to look-up the URLs of image packages to download.
Downloaded images are recorded in an inventory file that lives in the images install location.
This allows the downloader to skip images that were previously downloaded, and haven't changed
since.

Manifests are built into the downloader, but can also be accessed at uhd/images/manifest.txt.

Inventory files are JSON files called `inventory.json`, by default. It is possible to specify the
inventory file in command line arguments, but we don't recommend using this functionality unless
you're really sure you need it.
"""
import argparse
import hashlib
import json
import math
import os
import re
import shutil
import sys
import tempfile
import zipfile
import platform
from urllib.parse import urljoin  # Python 3

# For all the non-core-library imports, we will be extra paranoid and be very
# nice with error messages so that everyone understands what's up.
try:
    import requests
except ImportError:
    sys.stdout.write(
        "[ERROR] Missing module 'requests'! Please install it, e.g., by "
        "running 'pip install requests' or any other tool that can install "
        "Python modules.\n")
    if platform.system() == 'Windows':
        input('Hit Enter to continue.')
    exit(0)

# pylint: disable=bad-whitespace
_DEFAULT_TARGET_REGEX     = "(fpga|fw|windrv)_default"
_BASE_DIR_STRUCTURE_PARTS = ["share", "uhd", "images"]
_DEFAULT_INSTALL_PATH     = os.path.join("D:/bld/uhd_1600087591633/_h_env/Library", *_BASE_DIR_STRUCTURE_PARTS)
_DEFAULT_BASE_URL         = "https://files.ettus.com/binaries/cache/"
_INVENTORY_FILENAME       = "inventory.json"
_CONTACT                  = "support@ettus.com"
_DEFAULT_BUFFER_SIZE      = 8192
_DEFAULT_DOWNLOAD_LIMIT   = 100 * 1024 * 1024 # Bytes
_ARCHIVE_DEFAULT_TYPE     = "zip"
_UHD_VERSION              = "4.0.0.0-release"
_LOG_LEVELS = {
    "TRACE": 1,
    "DEBUG": 2,
    "INFO":  3,
    "WARN":  4,
    "ERROR": 5,
}
_LOG_LEVEL = _LOG_LEVELS["INFO"]
_YES = False
_PROXIES = {}
_BASE_URL_ENV_VAR = "UHD_IMAGES_URL"
# pylint: enable=bad-whitespace
# Note: _MANIFEST_CONTENTS are placed at the bottom of this file for aesthetic reasons

def log(level, message):
    """Logging function"""
    message_log_level = _LOG_LEVELS.get(level, 0)
    if message_log_level >= _LOG_LEVEL:
        sys.stderr.write(
            "[{level}] {message}\n".format(level=level, message=message))


def parse_args():
    """
    Setup argument parser and parse. Also does some sanity checks and sets some
    global variables we want to use.
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__)
    parser.add_argument('-t', '--types', action='append',
                        help="RegEx to select image sets from the manifest file.")
    parser.add_argument('-i', '--install-location', type=str, default=None,
                        help="Set custom install location for images")
    parser.add_argument('-m', '--manifest-location', type=str, default=None,
                        help="Set custom location for the manifest file")
    parser.add_argument('-I', '--inventory-location', type=str, default=None,
                        help="Set custom location for the inventory file")
    parser.add_argument('-l', '--list-targets', action="store_true", default=False,
                        help="Print targets in the manifest file to stdout, and exit.\n"
                        "To get relative paths only, specify an empty base URL (-b '').")
    parser.add_argument('--url-only', action="store_true", default=False,
                        help="With -l, only print the URLs, nothing else.")
    parser.add_argument("--buffer-size", type=int, default=_DEFAULT_BUFFER_SIZE,
                        help="Set download buffer size")
    parser.add_argument("--download-limit", type=int, default=_DEFAULT_DOWNLOAD_LIMIT,
                        help="Set threshold for download limits. Any download "
                             "larger than this will require approval, either "
                             "interactively, or by providing --yes.")
    parser.add_argument("--http-proxy", type=str,
                        help="Specify HTTP(S) proxy in the format "
                        "http[s]://user:pass@1.2.3.4:port\n"
                        "If this this option is not given, the environment "
                        "variables HTTP_PROXY/HTTPS_PROXY can also be used to specify a proxy.")
    parser.add_argument("-b", "--base-url", type=str,
                        help="Set base URL for images download location. "
                        "Defaults to ${} if set, or {} otherwise.".format(
                            _BASE_URL_ENV_VAR, _DEFAULT_BASE_URL))
    parser.add_argument("-k", "--keep", action="store_true", default=False,
                        help="Keep the downloaded images archives in the image directory")
    parser.add_argument("-T", "--test", action="store_true", default=False,
                        help="Verify the downloaded archives before extracting them")
    parser.add_argument("-y", "--yes", action="store_true", default=False,
                        help="Answer all questions with 'yes' (for scripting purposes).")
    parser.add_argument("-n", "--dry-run", action="store_true", default=False,
                        help="Print selected target without actually downloading them.")
    parser.add_argument("--refetch", action="store_true", default=False,
                        help="Ignore the inventory file and download all images.")
    parser.add_argument('-V', '--version', action='version', version=_UHD_VERSION)
    parser.add_argument('-q', '--quiet', action='count', default=0,
                        help="Decrease verbosity level")
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help="Increase verbosity level")
    args = parser.parse_args()
    # Set the verbosity
    global _LOG_LEVEL
    log("TRACE", "Default log level: {}".format(_LOG_LEVEL))
    _LOG_LEVEL = _LOG_LEVEL - args.verbose + args.quiet
    # Some sanitation that's easier to handle outside of the argparse framework:
    if not args.base_url:
        if os.environ.get(_BASE_URL_ENV_VAR):
            args.base_url = os.environ.get(_BASE_URL_ENV_VAR)
        else:
            args.base_url = _DEFAULT_BASE_URL
    log("INFO", "Using base URL: {}".format(args.base_url))
    if not args.base_url.endswith('/') and args.base_url != "":
        args.base_url += '/'
    if args.yes:
        global _YES
        _YES = True
    if args.http_proxy:
        global _PROXIES
        _PROXIES['http'] = args.http_proxy
        _PROXIES['https'] = args.http_proxy
    return args


def get_images_dir(args):
    """
    Figure out where to store the images.
    """
    if args.install_location:
        return args.install_location
    if os.environ.get("UHD_IMAGES_DIR"):
        log("DEBUG",
            "UHD_IMAGES_DIR environment variable is set, using to set "
            "install location.")
        return os.environ.get("UHD_IMAGES_DIR")
    return _DEFAULT_INSTALL_PATH


def ask_permission(question, default_no=True):
    """
    Ask the question, and have the user type y or n on the keyboard. If the
    global variable _YES is true, this always returns True without asking the
    question. Otherwise, return True if the answer is 'yes', 'y', or something
    similar.
    """
    if _YES:
        log("DEBUG", "Assuming the answer is 'yes' for this question: " +
            question)
        return True
    postfix = "[y/N]" if default_no else "[Y/n]"
    answer = input(question + " " + postfix)
    if answer and answer[0].lower() == 'y':
        return True
    return False


class TemporaryDirectory:
    """Class to create a temporary directory"""
    def __enter__(self):
        try:
            self.name = tempfile.mkdtemp()
            return self.name
        except Exception as ex:
            log("ERROR", "Failed to create a temporary directory (%s)" % ex)
            raise ex

    # Can return 'True' to suppress incoming exception
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            shutil.rmtree(self.name)
            log("TRACE", "Temp directory deleted.")
        except Exception as ex:
            log("ERROR", "Could not delete temporary directory: %s (%s)" % (self.name, ex))
        return exc_type is None


def get_manifest_raw(args):
    """
    Return the raw content of the manifest (i.e. the text file). It
    needs to be parsed to be of any practical use.
    """
    # If we're given a path to a manifest file, use it
    if args.manifest_location:
        manifest_fn = args.manifest_location
        if os.path.exists(manifest_fn):
            log("INFO", "Using manifest file at location: {}".format(manifest_fn))
            with open(manifest_fn, 'r') as manifest_file:
                manifest_raw = manifest_file.read()
        else:
            log("WARN", "Invalid manifest file specified: {}, using default"
                .format(manifest_fn))
            manifest_raw = _MANIFEST_CONTENTS       
    # Otherwise, use the CMake Magic manifest
    else:
        manifest_raw = _MANIFEST_CONTENTS
    log("TRACE", "Raw manifest contents: {}".format(manifest_raw))
    return manifest_raw


def parse_manifest(manifest_contents):
    """Parse the manifest file, returns a dictionary of potential targets"""
    manifest = {}
    for line in manifest_contents.split('\n'):
        line_unpacked = line.split()
        try:
            # Check that the line isn't empty or a comment
            if not line_unpacked or line.strip().startswith('#'):
                continue

            target, repo_hash, url, sha256_hash = line_unpacked
            manifest[target] = {"repo_hash": repo_hash,
                                "url": url,
                                "sha256_hash": sha256_hash,
                               }
        except ValueError:
            log("WARN", "Warning: Invalid line in manifest file:\n"
                "         {}".format(line))
            continue
    return manifest


def parse_inventory(inventory_fn):
    """Parse the inventory file, returns a dictionary of installed files"""
    try:
        if not os.path.exists(inventory_fn):
            log("INFO", "No inventory file found at {}. Creating an empty one.".format(inventory_fn))
            return {}
        with open(inventory_fn, 'r') as inventory_file:
            # TODO: verify the contents??
            return json.load(inventory_file)
    except Exception as ex:
        log("WARN", "Error parsing the inventory file. Assuming an empty inventory: {}".format(ex))
        return {}


def write_inventory(inventory, inventory_fn):
    """Writes the inventory to file"""
    try:
        with open(inventory_fn, 'w') as inventory_file:
            json.dump(inventory, inventory_file)
            return True
    except Exception as ex:
        log("ERROR", "Error writing the inventory file. Contents may be incomplete or corrupted.\n"
                     "Error message: {}".format(ex))
        return False


def lookup_urls(regex_l, manifest, inventory, refetch=False):
    """Takes a list of RegExs to match within the manifest, returns a list of tuples with
    (hash, URL) that match the targets and are not in the inventory"""
    selected_targets = []
    # Store whether or not we've found a target in the manifest that matches the requested type
    found_one = False
    for target in manifest.keys():
        # Iterate through the possible targets in the manifest.
        # If any of them match any of the RegExs supplied, add the URL to the
        # return list
        if all(map((lambda regex: re.findall(regex, target)), regex_l)):
            found_one = True
            log("TRACE", "Selected target: {}".format(target))
            target_info = manifest.get(target)
            target_url = target_info.get("url")
            target_hash = target_info.get("repo_hash")
            target_sha256 = target_info.get("sha256_hash")
            filename = os.path.basename(target_url)
            # Check if the same filename and hash appear in the inventory
            if not refetch and inventory.get(target, {}).get("repo_hash", "") == target_hash:
                # We already have this file, we don't need to download it again
                log("INFO", "Target {} is up to date.".format(target))
            else:
                # We don't have that exact file, add it to the list
                selected_targets.append({"target": target,
                                         "repo_hash": target_hash,
                                         "filename": filename,
                                         "url": target_url,
                                         "sha256_hash": target_sha256})
    if not found_one:
        log("INFO", "No targets matching '{}'".format(regex_l))
    return selected_targets


def print_target_list(manifest, args):
    """
    Print a list of targets.
    """
    char_offset = max(len(x) for x in manifest.keys())
    if not args.url_only:
        # Print a couple helpful lines,
        # then print each (Target, URL) pair in the manifest
        log("INFO", "Potential targets in manifest file:\n"
                    "{} : {}".format(
                        "# TARGET".ljust(char_offset),
                        "URL" if args.base_url else "RELATIVE_URL"))
        for key, value in sorted(manifest.items()):
            print("{target} : {base}{relpath}".format(
                target=key.ljust(char_offset),
                base=args.base_url,
                relpath=value["url"]))
    else:
        for manifest_item in manifest.items():
            print(args.base_url+manifest_item[1]["url"])


def download(
        images_url,
        filename,
        buffer_size=_DEFAULT_BUFFER_SIZE,
        print_progress=False,
        download_limit=None
    ):
    """ Run the download, show progress """
    download_limit = download_limit or _DEFAULT_DOWNLOAD_LIMIT
    log("TRACE", "Downloading {} to {}".format(images_url, filename))
    try:
        resp = requests.get(images_url, stream=True, proxies=_PROXIES,
                            headers={'User-Agent': 'UHD Images Downloader'})
    except TypeError:
        # requests library versions pre-4c3b9df6091b65d8c72763222bd5fdefb7231149
        # (Dec.'12) workaround
        resp = requests.get(images_url, prefetch=False, proxies=_PROXIES,
                            headers={'User-Agent': 'UHD Images Downloader'},
                            allow_redirects=True)
    if resp.status_code != 200:
        raise RuntimeError("URL does not exist: {}".format(images_url))
    filesize = float(resp.headers.get('content-length', -1))
    if filesize > download_limit:
        if not ask_permission(
                "The file size for this target ({:.1f} MiB) exceeds the "
                "download limit ({:.1f} MiB). Continue downloading?".format(
                    filesize/1024**2, download_limit/1024**2)):
            return 0, 0, ""
    if filesize == -1:
        if not ask_permission(
                "The file size for this target could not be determined. "
                "Continue downloading?"):
            return 0, 0, ""
    filesize_dl = 0
    base_filename = os.path.basename(filename)
    if print_progress and not sys.stdout.isatty():
        print_progress = False
        log("INFO", "Downloading {}, total size: {} kB".format(
            base_filename, filesize/1000 if filesize > 0 else "-unknown-"))
    with open(filename, "wb") as temp_file:
        sha256_sum = hashlib.sha256()
        for buff in resp.iter_content(chunk_size=buffer_size):
            if buff:
                temp_file.write(buff)
                filesize_dl += len(buff)
                sha256_sum.update(buff)
            if print_progress:
                status = r"%05d kB " % int(math.ceil(filesize_dl / 1000.))
                if filesize > 0:
                    status += r"/ %05d kB (%03d%%) " % (
                        int(math.ceil(filesize / 1000.)),
                        int(math.ceil(filesize_dl * 100.) / filesize),
                    )
                status += base_filename
                if os.name == "nt":
                    status += chr(8) * (len(status) + 1)
                else:
                    sys.stdout.write("\x1b[2K\r")  # Clear previous line
                sys.stdout.write(status)
                sys.stdout.flush()
    if print_progress:
        print('')
    if filesize <= 0:
        filesize = filesize_dl
    return filesize, filesize_dl, sha256_sum.hexdigest()


def delete_from_inv(target_info, inventory, images_dir):
    """
    Uses the inventory to delete the contents of the archive file specified by in `target_info`
    """
    target = inventory.get(target_info.get("target"), {})
    target_name = target.get("target")
    log("TRACE", "Removing contents of {} from inventory ({})".format(
        target, target.get("contents", [])))
    dirs_to_delete = []
    # Delete all of the files
    for image_fn in target.get("contents", []):
        image_path = os.path.join(images_dir, image_fn)
        if os.path.isfile(image_path):
            os.remove(image_path)
            log("TRACE", "Deleted {} from inventory".format(image_path))
        elif os.path.isdir(image_path):
            dirs_to_delete.append(image_fn)
        else: # File doesn't exist
            log("WARN", "File {} in inventory does not exist".format(image_path))
    # Then delete all of the (empty) directories
    for dir_path in dirs_to_delete:
        try:
            if os.path.isdir(dir_path):
                os.removedirs(dir_path)
        except os.error as ex:
            log("ERROR", "Failed to delete dir: {}".format(ex))
    inventory.pop(target_name, None)
    return True


def extract(archive_path, images_dir, test_zip=False):
    """
    Extract the contents of `archive_path` into `images_dir`

    Returns a list of files that were extracted.

    If test_zip is set, it will verify the zip file before extracting.
    """
    log("TRACE", "Attempting to extracted files from {}".format(archive_path))
    with zipfile.ZipFile(archive_path) as images_zip:
        # Check that the Zip file is valid, in which case `testzip()` returns
        # None.  If it's bad, that function will return a list of bad files
        try:
            if test_zip and images_zip.testzip():
                log("ERROR", "Could not extract the following invalid Zip file:"
                             " {}".format(archive_path))
                return []
        except OSError:
            log("ERROR", "Could not extract the following invalid Zip file:"
                         " {}".format(archive_path))
            return []
        images_zip.extractall(images_dir)
        archive_namelist = images_zip.namelist()
        log("TRACE", "Extracted files: {}".format(archive_namelist))
        return archive_namelist


def update_target(target_info, temp_dir, images_dir, inventory, args):
    """
    Handle the updating of a single target.
    """
    target_name = target_info.get("target")
    target_sha256 = target_info.get("sha256_hash")
    filename = target_info.get("filename")
    temp_path = os.path.join(temp_dir, filename)
    # Add a trailing slash to make sure that urljoin handles things properly
    full_url = urljoin(args.base_url+'/', target_info.get("url"))
    _, downloaded_size, downloaded_sha256 = download(
        images_url=full_url,
        filename=temp_path,
        buffer_size=args.buffer_size,
        print_progress=(_LOG_LEVEL <= _LOG_LEVELS.get("INFO", 3))
    )
    if downloaded_size == 0:
        log("INFO", "Skipping target: {}".format(target_name))
        return
    log("TRACE", "{} successfully downloaded ({} Bytes)"
        .format(temp_path, downloaded_size))
    # If the SHA256 in the manifest has the value '0', this is a special case
    # and we just skip the verification step
    if target_sha256 == '0':
        log("DEBUG", "Skipping SHA256 check for {}.".format(full_url))
    # If the check fails, print an error and don't unzip the file
    elif downloaded_sha256 != target_sha256:
        log("ERROR", "Downloaded SHA256 does not match manifest for {}!"
            .format(full_url))
        return
        # Note: this skips the --keep option, so we'll never keep image packages
        #       that fail the SHA256 checksum
    ## Now copy the contents to the final destination (the images directory)
    delete_from_inv(target_info, inventory, images_dir)
    if os.path.splitext(temp_path)[1].lower() == '.zip':
        archive_namelist = extract(
            temp_path,
            images_dir,
            args.test)
        if args.keep:
            # If the user wants to keep the downloaded archive,
            # save it to the images directory and add it to the inventory
            shutil.copy(temp_path, images_dir)
            archive_namelist.append(filename)
    else:
        archive_namelist = []
        shutil.copy(temp_path, images_dir)
    ## Update inventory
    inventory[target_name] = {"repo_hash": target_info.get("repo_hash"),
                              "contents": archive_namelist,
                              "filename": filename}


def main():
    """
    Main function; does whatever the user requested (download or list files).

    Returns True on successful execution.
    """
    args = parse_args()
    images_dir = get_images_dir(args)
    log("INFO", "Images destination: {}".format(os.path.abspath(images_dir)))
    try:
        manifest = parse_manifest(get_manifest_raw(args))
        if args.list_targets:
            print_target_list(
                manifest,
                args
            )
            return True
        log("TRACE", "Manifest:\n{}".format(
            "\n".join("{}".format(item) for item in manifest.items())
        ))

        # Strip down path until we find first parent directory of images_dir
        # that exists and then check if we have proper write permissions
        images_dir_parent = images_dir
        while not os.path.exists(os.path.abspath(images_dir_parent)):
            path_pair = os.path.split(images_dir_parent)
            if images_dir_parent == path_pair[0]:
                # Prevent infinite loop
                break
            images_dir_parent = path_pair[0]
        if not os.access(os.path.abspath(images_dir_parent), os.W_OK):
            log("ERROR", "Invalid permissions to write images destination")
            return False

        # Read the inventory into a dictionary we can perform lookups on
        default_inventory_fn = os.path.join(images_dir, _INVENTORY_FILENAME)
        if args.inventory_location:
            if os.path.isfile(args.inventory_location):
                inventory_fn = args.inventory_location
            else:
                log("WARN", "Invalid inentory file specified: {}, using default: {}"
                    .format(args.inventory_location, default_inventory_fn))
                inventory_fn = default_inventory_fn
        else:
            inventory_fn = default_inventory_fn
        inventory = parse_inventory(inventory_fn=inventory_fn)
        log("TRACE", "Inventory: {}\n{}".format(
            os.path.abspath(inventory_fn),
            "\n".join("{}".format(item) for item in inventory.items())
        ))

        # Determine the URLs to download based on the input regular expressions
        if not args.types:
            types_regex_l = [_DEFAULT_TARGET_REGEX]
        else:
            types_regex_l = args.types

        log("TRACE", "RegExs for target selection: {}".format(types_regex_l))
        targets_info = lookup_urls(types_regex_l, manifest, inventory, args.refetch)
        # Exit early if we don't have anything to download
        if targets_info:
            target_urls = [info.get("url") for info in targets_info]
            log("DEBUG", "URLs to download:\n{}".format(
                "\n".join("{}".format(item) for item in target_urls)
            ))
        else:
            return True

        ## Now download all the images archives into a temp directory
        if args.dry_run:
            for target_info in targets_info:
                log("INFO", "[Dry Run] Fetch target: {}".format(
                    target_info.get("filename")))
            return True
        with TemporaryDirectory() as temp_dir:
            for target_info in targets_info:
                update_target(
                    target_info,
                    temp_dir,
                    images_dir,
                    inventory,
                    args
                )
        ## Update inventory with all the new content
        write_inventory(inventory, inventory_fn)

    except Exception as ex:
        log("ERROR", "Downloader raised an unhandled exception: {ex}\n"
            "You can run this again with the '--verbose' flag to see more information\n"
            "If the problem persists, please email the output to: {contact}"
            .format(contact=_CONTACT, ex=ex))
        # Again, we wait on Windows systems because if this is executed in a
        # window, and immediately fails, the user doesn't have a way to see the
        # error message, and if they're not very savvy, they won't know how to
        # execute this in a shell.
        if not _YES and platform.system() == 'Windows':
            input('Hit Enter to continue.')
        return False
    log("INFO", "Images download complete.")
    return True

# Placing this near the end of the file so we don't clutter the top
_MANIFEST_CONTENTS = """# UHD Image Manifest File
# Target    hash    url     SHA256
# X300-Series
x3xx_x310_fpga_default          uhd-be53058       x3xx/uhd-be53058/x3xx_x310_fpga_default-gbe53058.zip                  baa6082515704090ce701cb89dddcc2c99b42fcdb72648f0d465520aa11cfc8b
x3xx_x300_fpga_default          uhd-be53058       x3xx/uhd-be53058/x3xx_x300_fpga_default-gbe53058.zip                  3d58e95e1e53716eec3f62067967a2baedeb8f2d1f3cfe4ecb9d01f984140291
# Example daughterboard targets (none currently exist)
#x3xx_twinrx_cpld_default   example_target
#dboard_ubx_cpld_default    example_target

# E-Series
e3xx_e310_sg1_fpga_default      uhd-be53058       e3xx/uhd-be53058/e3xx_e310_sg1_fpga_default-gbe53058.zip             b708d89d2d58fd8d8f180f98d08bba177c117d8b0b8251bfcfb7dd4e4657d165
e3xx_e310_sg3_fpga_default      uhd-be53058       e3xx/uhd-be53058/e3xx_e310_sg3_fpga_default-gbe53058.zip             6a57ffa57dc29138185f6e9c208a886c15d62c9cd838d41cda9349f5d13c082d
e3xx_e320_fpga_default          uhd-413ebee       e3xx/uhd-413ebee/e3xx_e320_fpga_default-g413ebee.zip                 296c8d86b0f5feb3b551c2c5f856e221cdf7288eeacd16c8e1c56c0bfa0b4934

# E310 Filesystems
e3xx_e310_sdk_default         meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e310_sdk_default-v4.0.0.0.zip       0
e3xx_e310_sg1_mender_default  meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e310_sg1_mender_default-v4.0.0.0.zip    0
e3xx_e310_sg1_sdimg_default   meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e310_sg1_sdimg_default-v4.0.0.0.zip     0
e3xx_e310_sg3_mender_default  meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e310_sg3_mender_default-v4.0.0.0.zip    0
e3xx_e310_sg3_sdimg_default   meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e310_sg3_sdimg_default-v4.0.0.0.zip     0

# E320 Filesystems, etc
e3xx_e320_sdk_default         meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e320_sdk_default-v4.0.0.0.zip       0
e3xx_e320_mender_default      meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e320_mender_default-v4.0.0.0.zip    0
e3xx_e320_sdimg_default       meta-ettus-v4.0.0.0  e3xx/meta-ettus-v4.0.0.0/e3xx_e320_sdimg_default-v4.0.0.0.zip     0

# N300-Series
n3xx_n310_fpga_default          uhd-be53058       n3xx/uhd-be53058/n3xx_n310_fpga_default-gbe53058.zip                 edc7e2e870446efaf3df75433913215b2ce81965e35edacaea1b1ae253325fd6
n3xx_n300_fpga_default          uhd-be53058       n3xx/uhd-be53058/n3xx_n300_fpga_default-gbe53058.zip                 6f9503b4a3e0f006a46eb131f0a9e4decdcb8dcdbe2416bb7b727f7f518950c3
n3xx_n320_fpga_default          uhd-be53058        n3xx/uhd-be53058/n3xx_n320_fpga_default-gbe53058.zip                88e5c504a5ddcb63c4b0bd439951b6b4166bf4ac2cbca1cdb441f0e9b030acb4
n3xx_n310_cpld_default          fpga-6bea23d        n3xx/fpga-6bea23d/n3xx_n310_cpld_default-g6bea23d.zip                ef128dcd265ee8615b673021d4ee84c39357012ffe8b28c8ad7f893f9dcb94cb
n3xx_n320_cpld_default          fpga-4bc2c6f       n3xx/fpga-4bc2c6f/n3xx_n320_cpld_default-g4bc2c6f.zip              6680a9363efc5fa8b5a68beb3dff44f2e314b94e716e3a1751aba0fed1f384da
# N3XX Mykonos firmware
#n3xx_n310_fw_default           fpga-6bea23d        n3xx/fpga-6bea23d/n3xx_n310_fw_default-g6bea23d.zip                     0
# N300-Series Filesystems, etc
n3xx_common_sdk_default         meta-ettus-v4.0.0.0  n3xx/meta-ettus-v4.0.0.0/n3xx_common_sdk_default-v4.0.0.0.zip       0
n3xx_common_mender_default      meta-ettus-v4.0.0.0  n3xx/meta-ettus-v4.0.0.0/n3xx_common_mender_default-v4.0.0.0.zip    0
n3xx_common_sdimg_default       meta-ettus-v4.0.0.0  n3xx/meta-ettus-v4.0.0.0/n3xx_common_sdimg_default-v4.0.0.0.zip     0

# B200-Series
b2xx_b200_fpga_default          uhd-f2ec5c9       b2xx/uhd-f2ec5c9/b2xx_b200_fpga_default-gf2ec5c9.zip               026c3b2613ceabbe7f47e479c2933e41ca054e7b1f85a119d86d97b6df4ca744
b2xx_b200mini_fpga_default      uhd-f2ec5c9       b2xx/uhd-f2ec5c9/b2xx_b200mini_fpga_default-gf2ec5c9.zip           3c122145a5bd23dc50ac35b9d18dfe3f3224eaed4242274765ada89c55838ce4
b2xx_b210_fpga_default          uhd-f2ec5c9       b2xx/uhd-f2ec5c9/b2xx_b210_fpga_default-gf2ec5c9.zip               4eba82a362369cc21ba679a4debdcd6aabbc1ecea295eb4aefd94f2ba5e6ed70
b2xx_b205mini_fpga_default      uhd-f2ec5c9       b2xx/uhd-f2ec5c9/b2xx_b205mini_fpga_default-gf2ec5c9.zip           a38fcecabcef43d57e463aa50f0b07ba21d7e9211428fbc5ef47bfeb75542e93
b2xx_common_fw_default          uhd-437381b         b2xx/uhd-437381b/b2xx_common_fw_default-g437381b.zip                  a71b01cb7dfea975f26516b5b18f9d2bebd5083effaf21a79fd5925d2aaf6a32

# USRP2 Devices
usrp2_usrp2_fw_default          fpga-6bea23d        usrp2/fpga-6bea23d/usrp2_usrp2_fw_default-g6bea23d.zip                  d523a18318cb6a7637be40484bf03a6f54766410fee2c1a1f72e8971ea9a9cb6
usrp2_usrp2_fpga_default        fpga-6bea23d        usrp2/fpga-6bea23d/usrp2_usrp2_fpga_default-g6bea23d.zip                505c70aedc8cdfbbfe654bcdbe1ce604c376e733a44cdd1351571f61a7f1cb49
usrp2_n200_fpga_default         fpga-6bea23d        usrp2/fpga-6bea23d/usrp2_n200_fpga_default-g6bea23d.zip                 833a0098d66c0c502b9c3975d651a79e125133c507f9f4b2c472f9eb96fdaef8
usrp2_n200_fw_default           fpga-6bea23d        usrp2/fpga-6bea23d/usrp2_n200_fw_default-g6bea23d.zip                   3eee2a6195caafe814912167fccf2dfc369f706446f8ecee36e97d2c0830116f
usrp2_n210_fpga_default         fpga-6bea23d        usrp2/fpga-6bea23d/usrp2_n210_fpga_default-g6bea23d.zip                 5ce68ac539ee6eeb7d04fb3127c1fabcaff442a8edfaaa2f3746590f9df909bd
usrp2_n210_fw_default           fpga-6bea23d        usrp2/fpga-6bea23d/usrp2_n210_fw_default-g6bea23d.zip                   3646fcd3fc974d18c621cb10dfe97c4dad6d282036dc63b7379995dfad95fb98
n230_n230_fpga_default          fpga-e57dfe0        n230/fpga-e57dfe0/n230_n230_fpga_default-ge57dfe0.zip                   f115b78a1893da0c2a481257351b834721f8a5468dfae6a34e3a68a3bc0065cd

# USRP1 Devices
usrp1_usrp1_fpga_default        fpga-6bea23d        usrp1/fpga-6bea23d/usrp1_usrp1_fpga_default-g6bea23d.zip                03bf72868c900dd0853bf48e2ede91058d579829b0e70c021e51b0e282d1d5be
usrp1_b100_fpga_default         fpga-6bea23d        usrp1/fpga-6bea23d/usrp1_b100_fpga_default-g6bea23d.zip                 7f2306f21e17aa3fae3f966d08c6297d6cf42041974f846ca89f0d633ece8769
usrp1_b100_fw_default           fpga-6bea23d        usrp1/fpga-6bea23d/usrp1_b100_fw_default-g6bea23d.zip                   867f17fac085535dbcb01c226ce87acf49806de6ed0ae9b214d7c8da86e2a71d

# Octoclock
octoclock_octoclock_fw_default  uhd-14000041        octoclock/uhd-14000041/octoclock_octoclock_fw_default-g14000041.zip     8da7f1af8cecb7f6259a237a18c39058ba69a11567fa373cffc9704031a1d053

# Legacy USB Windows drivers
usb_common_windrv_default       uhd-14000041        usb/uhd-14000041/usb_common_windrv_default-g14000041.zip                835e94b2bdf2312fd3881a1b78e2ec236c1f42b7a5bd3927f85f73cf5e3a5231
"""
if __name__ == "__main__":
    sys.exit(not main())



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
UHD Power Cal: USRP Calbration Utility Objects
"""

import time
import inspect
import sys
import numpy
import uhd

from . import database
from .tone_gen import ToneGenerator

NUM_SAMPS_PER_EST = int(1e6)
# Limits for the power estimation algorithm. For good estimates, we want the
# signal to be at -6 dBFS, but not outside of an upper or lower limit.
PWR_EST_LLIM = -20
PWR_EST_IDEAL_LEVEL = -6
PWR_EST_ULIM = -3
SIGPWR_LOCK_MAX_ITER = 4

# The default distance between frequencies at which we measure
DEFAULT_FREQ_STEP = 10e6 # Hz
DEFAULT_SAMP_RATE = 5e6

def get_streamer(usrp, direction, chan):
    """
    Create an appropriate streamer object for this channel
    """
    stream_args = uhd.usrp.StreamArgs('fc32', 'sc16')
    stream_args.channels = [chan]
    return usrp.get_rx_stream(stream_args) if direction == 'rx' \
           else usrp.get_tx_stream(stream_args)

def get_default_gains(direction, gain_range, gain_step):
    """
    Create a equidistant gain range for calibration
    """
    assert direction in ('rx', 'tx')
    if direction == 'tx':
        return numpy.arange(0, gain_range.stop(), gain_step)
    return numpy.arange(gain_range.stop(), 0, -gain_step)

def get_usrp_power(streamer, num_samps=NUM_SAMPS_PER_EST, chan=0):
    """
    Return the measured input power in dBFS

    The return value is a list of dBFS power values, one per channel.
    """
    recv_buffer = numpy.zeros(
        (streamer.get_num_channels(), num_samps), dtype=numpy.complex64)
    metadata = uhd.types.RXMetadata()
    stream_cmd = uhd.types.StreamCMD(uhd.types.StreamMode.num_done)
    stream_cmd.num_samps = num_samps
    stream_cmd.stream_now = True
    streamer.issue_stream_cmd(stream_cmd)
    # Pass in long timeout, so we can rx the entire buffer in one go
    samps_recvd = streamer.recv(recv_buffer, metadata, 5.0)
    if samps_recvd != num_samps:
        raise RuntimeError(
            "ERROR! get_usrp_power(): Did not receive the correct number of samples!")
    return uhd.dsp.signals.get_power_dbfs(recv_buffer[chan])


def subtract_power(p1_db, p2_db):
    """
    Return the power of p1 subtracted from p2, where both are given in logarithmic
    units.
    """
    return 10 * numpy.log10(10**(p1_db/10) - 10**(p2_db/10))

###############################################################################
# Base Class
###############################################################################
class USRPCalibratorBase(object):
    """
    Base class for device calibration. Every USRP that can do power calibration
    needs to implement this.
    """
    # These should be overriden to device-specific values
    max_input_power = -20 # -20 happens to be a safe value for all devices
    min_detectable_signal = -70
    default_rate = DEFAULT_SAMP_RATE
    lo_offset = 0.0

    def __init__(self, usrp, meas_dev, direction, **kwargs):
        self._usrp = usrp
        self._meas_dev = meas_dev
        # This makes sure our measurement will not destroy the DUT
        self._meas_dev.max_output_power = self.max_input_power
        self._dir = direction
        self._desired_gain_step = kwargs.get('gain_step', 10)
        self._id = usrp.get_usrp_rx_info(0).get('mboard_id')
        self._mb_serial = usrp.get_usrp_rx_info(0).get('mboard_serial')
        # Littler helper to print stuff with a device ID prefix.
        self.log = lambda *args, **kwargs: print("[{}]".format(self._id), *args, **kwargs)
        # Channel, antenna, and streamer will get updated in update_port()
        self._chan = None
        self._ant = ""
        self._streamer = None
        # These dictionaries store the results that get written out as well as
        # the noise floor for reference
        self.results = {} # This must be of the form results[freq][gain] = power
        self._noise = {}
        # The tone generator object is only needed for Tx measurements, and is
        # initialized conditionaly in init()
        self._tone_gen = None
        # The gains can be overridden by the device if non-equidistant gains are
        # desired. However, gains must be increasing order for Tx measurements,
        # and in decreasing order for Rx measurements.
        self._gains = get_default_gains(
            direction,
            getattr(self._usrp, 'get_{}_gain_range'.format(self._dir))(),
            self._desired_gain_step)
        # You might want to override this, but it's not important. It will
        # become the 'name' argument for the power cal factory.
        self.cal_name = self._id + " Power Cal"
        # The child class can store temperature and ref gain here
        self.temp = None
        self.ref_gain = None

    def init(self, rate, tone_freq, amplitude):
        """
        Initialize device with finalized values. Not that __init__() needs to
        finish before the call site knows the rate, so we can' fold this into
        __init__().
        """
        if self._dir == 'tx':
            self._tone_gen = ToneGenerator(rate, tone_freq, amplitude)

    def update_port(self, chan, antenna):
        """
        Notify the device that we've switched channel and/or antenna.
        """
        self.log("Switching to channel {}, antenna {}.".format(chan, antenna))
        self._ant = antenna
        if chan != self._chan:
            # This will be an RX streamer for RX power cal, and a TX streamer
            # for TX power cal.
            self._streamer = get_streamer(self._usrp, self._dir, chan)
            if self._dir == 'tx':
                self._tone_gen.set_streamer(self._streamer)
        self._chan = chan

    def _get_frequencies(self, start_hint=None, stop_hint=None, step_hint=None):
        """
        Return an iterable of frequencies for testing.

        The default will check the hints against the device, but otherwise heed
        them.

        If a particular device needs to check specific frequencies, then
        override this.
        """
        start_min = \
            getattr(self._usrp, 'get_{}_freq_range'.format(self._dir))(
                self._chan).start()
        start_hint = start_hint or start_min
        start = max(start_hint, start_min)
        stop_max = \
            getattr(self._usrp, 'get_{}_freq_range'.format(self._dir))(
                self._chan).stop()
        stop_hint = stop_hint or stop_max
        stop = min(stop_hint, stop_max)
        step = step_hint or DEFAULT_FREQ_STEP
        return numpy.arange(start, stop + step, step)

    def init_frequencies(self, start_hint, stop_hint, step_hint):
        """
        Return an iterable of frequencies for testing.

        The default will check the hints against the device, but otherwise heed
        them.

        Then it will measure the noise floor across frequency to get a good
        baseline measurement.
        """
        freqs = self._get_frequencies(start_hint, stop_hint, step_hint)
        if self._dir == 'tx':
            print("===== Measuring noise floor across frequency...")
            for freq in freqs:
                self._meas_dev.set_frequency(freq)
                self._noise[freq] = self._meas_dev.get_power()
                print("[TX] Noise floor: {:2} MHz => {:+6.2f} dBm"
                      .format(freq/1e6, self._noise[freq]))
        else: # Rx
            print("===== Measuring noise floor across frequency and gain...")
            for freq in freqs:
                self._noise[freq] = {}
                tune_req = uhd.types.TuneRequest(freq)
                self._usrp.set_rx_freq(tune_req, self._chan)
                for gain in self._gains:
                    self._usrp.set_rx_gain(gain, self._chan)
                    self._noise[freq][gain] = get_usrp_power(self._streamer)
                    print("[RX] Noise floor: {:2} MHz / {} dB => {:+6.2f} dBFS"
                          .format(freq/1e6, gain, self._noise[freq][gain]))
        return freqs

    def start(self):
        """
        Initialize the device for calibration
        """
        if self._dir == 'tx':
            self._tone_gen.start()
        else:
            self._meas_dev.enable(True)

    def stop(self, store=True):
        """
        Shut down the device after calibration
        """
        if self._dir == 'tx':
            self._tone_gen.stop()
        else:
            self._meas_dev.enable(False)
        if store:
            self.store()

    def run_rx_cal(self, freq):
        """
        Run the actual RX calibration for this frequency.
        """
        # Go to highest gain, lock in signal generator
        self._usrp.set_rx_gain(max(self._gains), self._chan)
        time.sleep(0.1) # Settling time of the USRP, highly conservative
        self.log("Locking in signal generator power...")
        self.log("Requesting input power: {:+.2f} dBm."
                 .format(self.min_detectable_signal))
        usrp_input_power = self._meas_dev.set_power(self.min_detectable_signal)
        recvd_power = get_usrp_power(self._streamer)
        self.log("Got input power: {:+.2f} dBm. Received power: {:.2f} dBFS. "
                 "Requesting new input power: {:+.2f} dBm."
                 .format(usrp_input_power,
                         recvd_power,
                         usrp_input_power + PWR_EST_IDEAL_LEVEL - recvd_power))
        usrp_input_power = self._meas_dev.set_power(
            usrp_input_power + PWR_EST_IDEAL_LEVEL - recvd_power)
        siggen_locked = False
        for _ in range(SIGPWR_LOCK_MAX_ITER):
            recvd_power = get_usrp_power(self._streamer)
            if PWR_EST_LLIM <= recvd_power <= PWR_EST_ULIM:
                siggen_locked = True
                break
            self.log("Receiving input power: {:+.2f} dBFS.".format(recvd_power))
            power_delta = PWR_EST_IDEAL_LEVEL - recvd_power
            # Update power output by the delta from the desired input value:
            self.log("Requesting input power: {:+.2f} dBm."
                     .format(usrp_input_power + power_delta))
            usrp_input_power = self._meas_dev.set_power(usrp_input_power + power_delta)
        if not siggen_locked:
            raise RuntimeError(
                "Unable to lock siggen within {} iterations! Last input power level: {:+6.2f} dBm."
                .format(SIGPWR_LOCK_MAX_ITER, usrp_input_power))
        self.log("Locked signal generator in at input power level: {:+6.2f} dBm."
                 .format(usrp_input_power))
        # Now iterate through gains
        results = {}
        # Gains are in decreasing order!
        last_gain = self._gains[0]
        for gain in self._gains:
            self._usrp.set_rx_gain(gain, self._chan) # Set the new gain
            self.log("Set gain to: {} dB. Got gain: {} dB."
                     .format(gain, self._usrp.get_rx_gain(self._chan)))
            time.sleep(0.1) # Settling time of the USRP, highly conservative
            gain_delta = last_gain - gain # This is our gain step
            if gain_delta:
                # If we decrease the device gain, we need to crank up the input
                # power
                usrp_input_power = self._meas_dev.set_power(
                    min(usrp_input_power + gain_delta, self.max_input_power))
                # usrp_input_power = self._meas_dev.set_power(usrp_input_power + gain_delta)
                self.log("New input power is: {:+.2f} dBm".format(usrp_input_power))
            recvd_power = get_usrp_power(self._streamer)
            self.log("Received power: {:.2f} dBFS".format(recvd_power))
            # It's possible that we lose the lock on the signal power, so allow
            # for a correction
            if not PWR_EST_LLIM <= recvd_power <= PWR_EST_ULIM:
                power_delta = PWR_EST_IDEAL_LEVEL - recvd_power
                self.log("Adapting input power to: {:+.2f} dBm."
                         .format(usrp_input_power + power_delta))
                usrp_input_power = self._meas_dev.set_power(usrp_input_power + power_delta)
                self.log("New input power is: {:+.2f} dBm".format(usrp_input_power))
                # And then of course, measure again
                recvd_power = get_usrp_power(self._streamer)
                self.log("Received power: {:.2f} dBFS".format(recvd_power))
            # Note: The noise power should be way down there, and really
            # shouldn't matter. We subtract it anyway for formal correctness.
            recvd_signal_power = subtract_power(recvd_power, self._noise[freq][gain])
            # A note on the following equation: 'recvd_signal_power' is in dBFS,
            # and usrp_input_power is in dBm. However, this is the reference
            # signal, so we need the power (in dBm) that corresponds to 0 dBFS.
            # The assumption is that digital gain is linear, so what we really
            # want is usrp_input_power - (recvd_signal_power - 0dBFS), and the
            # result of the equation is in dBm again. We omit the subtract-by-zero
            # since our variables don't have units.
            results[gain] = usrp_input_power - recvd_signal_power
            self.log("{:2} dB => {:+6.2f} dBm".format(gain, results[gain]))
            # If we get too close to the noise floor, we stop
            if recvd_power - self._noise[freq][gain] <= 1.5:
                self.log("Can no longer detect input signal. Terminating.")
                break
        self.results[freq] = results

    def run_tx_cal(self, freq):
        """
        Run the actual TX calibration for this frequency.
        """
        results = {}
        for gain in self._gains:
            self._usrp.set_tx_gain(gain, self._chan)
            time.sleep(0.1) # Settling time of the USRP, highly conservative
            results[gain] = self._meas_dev.get_power()
            self.log("{:2} dB => {:+6.2f} dBm".format(gain, results[gain]))
        self.results[freq] = results

    def store(self):
        """
        Return the results object
        """
        chan_info = getattr(self._usrp, "get_usrp_{}_info".format(self._dir))(self._chan)
        cal_key = chan_info.get("{}_ref_power_key".format(self._dir))
        cal_serial = chan_info.get("{}_ref_power_serial".format(self._dir))
        cal_data = uhd.usrp.cal.PwrCal(self.cal_name, cal_serial, int(time.time()))
        if self.temp:
            cal_data.set_temperature(self.temp)
        if self.ref_gain:
            cal_data.set_ref_gain(self.ref_gain)
        for freq, results in self.results.items():
            max_power = max(results.values())
            min_power = min(results.values())
            cal_data.add_power_table(results, min_power, max_power, freq)
        database.write_cal_data(
            cal_key,
            cal_serial,
            cal_data.serialize())
        self.results = {}

class B200Calibrator(USRPCalibratorBase):
    """
    B200 calibration
    """
    mboard_ids = ('B200', 'B210', 'B200mini', 'B205mini')
    # Choosing 5 MHz: It is a small rate, but carries enough bandwidth to receive
    # a tone. By default, the auto MCR will kick in and set the MCR to 40 Msps,
    # thus engaging all halfbands.
    default_rate = 5e6
    # Choosing an LO offset of 10 MHz: At 5 Msps, the LO will never be within
    # our estimate. B200 generally has good DC offset / IQ balance performance,
    # but we still try and avoid DC as much as possible.
    lo_offset = 10e6

    def __init__(self, usrp, meas_dev, direction, **kwargs):
        super().__init__(usrp, meas_dev, direction, **kwargs)
        print("===== Reading temperature... ", end="")
        self.temp = self._usrp.get_rx_sensor("temp").to_int()
        print("{} C".format(self.temp))
        # TODO don't hard code
        self.ref_gain = 60

class X300Calibrator(USRPCalibratorBase):
    """
    X300/X310 calibration

    Notes / TODOs:
    - The X310 must avoid frequencies that are multiples of 200 MHz; here be
      harmonics (TODO)
    - TwinRX needs its own special method. It needs to be run once with one
      channel in the streamer, and once with two channels in the streamer. Or we
      come up with a clever way of enabling the other channel without modifying
      the streamer. A poke to the prop tree might do the trick.
    """
    mboard_ids = ('X300', 'X310', 'NI-2974')
    # Choosing 5 MHz: It is a small rate, but carries enough bandwidth to receive
    # a tone. It's 1/40 the master clock rate, which means it'll engage max
    # halfbands.
    default_rate = 5e6
    # Choosing an LO offset of 10 MHz: At 5 Msps, the LO will never be within
    # our estimate, so it doesn't matter if this device is DC offset / IQ balance
    # calibrated.
    lo_offset = 10e6


###############################################################################
# The dispatch function
###############################################################################
def get_usrp_calibrator(usrp, meas_dev, direction, **kwargs):
    """
    Return a USRP calibrator object.
    """
    usrp_type = \
        getattr(usrp, 'get_usrp_{}_info'.format(direction))().get('mboard_id')
    if usrp_type is None:
        raise RuntimeError("Could not determine USRP type!")
    print("=== Detected USRP type:", usrp_type)
    for _, obj in inspect.getmembers(sys.modules[__name__]):
        try:
            if issubclass(obj, USRPCalibratorBase) \
                    and usrp_type in getattr(obj, 'mboard_ids', ''):
                return obj(usrp, meas_dev, direction, **kwargs)
        except TypeError:
            continue
    raise RuntimeError("No USRP calibrator object found for device type: {}"
                       .format(usrp_type))



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Provide a tone generator class for USRPs.
"""

import threading
import numpy
import uhd

class ToneGenerator:
    """
    Class that can output a tone from a different thread until told to stop
    """
    def __init__(self, rate, freq, ampl, streamer=None):
        self._streamer = streamer
        self._buffer = uhd.dsp.signals.get_continuous_tone(rate, freq, ampl)
        self._run = False
        self._thread = None

    def set_streamer(self, streamer):
        """
        Update streamer object
        """
        if self._run:
            self.stop()
        self._streamer = streamer

    def start(self):
        """
        Spawn the thread in the background
        """
        if not self._streamer:
            raise RuntimeError("No streamer defined!")
        self._run = True
        self._thread = threading.Thread(target=self._worker)
        self._thread.start()
        self._thread.setName("cal_tx")

    def stop(self):
        """
        Stop the transmitter
        """
        self._run = False
        self._thread.join()
        self._thread = None

    def _worker(self):
        """ Here is where the action happens """
        metadata = uhd.types.TXMetadata()
        while self._run:
            # Give it a long-ish timeout so we don't have to throttle in here
            if self._streamer.send(self._buffer, metadata, 1.0) != len(self._buffer):
                print("WARNING: Failed to transmit entire buffer in ToneGenerator!")
        # Send an EOB packet with a single zero-valued sample to close out TX
        metadata.end_of_burst = True
        self._streamer.send(
            numpy.array([0], dtype=numpy.complex64), metadata, 0.1)



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
VISA-type measurement devices
"""

import re
import sys
import inspect
import importlib

class VISADevice:
    """
    Parent class for VISA/SCPI devices (that can be accessed via PyVISA)
    """
    # Must contain a dictionary res_ids. See class below as an example

    def __init__(self, resource):
        self.res = resource
        # You can query the exact model etc. like this:
        # id = self._res.query("*IDN?").strip(

    def init_power_meter(self):
        """
        Initialize this device to measure power.
        """
        raise NotImplementedError()

    def init_signal_generator(self):
        """
        Initialize this device to generate signals.
        """
        raise NotImplementedError()

    def set_frequency(self, freq):
        """
        Set frequency
        """
        raise NotImplementedError()

    def get_power_dbm(self):
        """
        Return the received/measured power in dBm (power meter) or return  the
        output power it's currently set to (signal generator).
        """
        raise NotImplementedError()


class USBPowerMeter(VISADevice):
    """
    USB-based power measurement devices
    """
    # The keys of res_ids are regular expressions, which have to match the
    # resource ID of the VISA device. The values are a name for the device, which
    # is used for informing the user which driver was found.
    #
    # A class can match any number of resource IDs. If the commands used depend
    # on the specific ID, it can be queried in the appropriate init function
    # using the *IDN? command which is understood by all VISA devices.
    res_ids = {
        r'USB\d+::2733::376::\d+::0::INSTR': 'R&S NRP-6A',
    }

    def init_power_meter(self):
        """
        Enable the sensor to read power
        """
        self.res.timeout = 5000
        self.res.write("SENS:AVER:COUN 20")
        self.res.write("SENS:AVER:COUN:AUTO ON")
        self.res.write('UNIT:POW DBM')
        self.res.write('SENS:FUNC "POW:AVG"')

    def init_signal_generator(self):
        """
        This class is for power meters, so no bueno
        """
        raise RuntimeError("This device cannot be used for signal generation!")

    def set_frequency(self, freq):
        """
        Set frequency
        """
        self.res.write('SENS:FREQ {}'.format(freq))

    def get_power_dbm(self):
        """
        Return measured power in dBm
        """
        self.res.write('INIT:IMM')
        return float(self.res.query('FETCH?'))

###############################################################################
# The dispatch function
###############################################################################
def get_visa_device(resource, key, opt_dict):
    """
    Return the VISA device object
    """
    def match_res(obj):
        """
        Check if a class obj matches the requested key
        """
        for pattern, res_id in getattr(obj, 'res_ids', {}).items():
            if re.match(pattern, key):
                print("Found VISA device: {}".format(res_id))
                return True
        return False
    # Import additional modules if requested
    members = inspect.getmembers(sys.modules[__name__])
    if 'import' in opt_dict:
        try:
            print("Loading external module: {}".format(opt_dict.get('import')))
            external_module = importlib.import_module(opt_dict.get('import'))
            members += inspect.getmembers(external_module)
        except (ModuleNotFoundError, ImportError):
            print("WARNING: Could not import module '{}'"
                  .format(opt_dict.get('import')))
    # Now browse classes and find one that matches
    for _, obj in members:
        try:
            if issubclass(obj, VISADevice) and match_res(obj):
                return obj(resource)
        except TypeError:
            continue
    raise RuntimeError("No VISA device class found for key: {}".format(key))



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Utilities for generating/analyzing signals
"""

import math
import numpy
import uhd

def get_continuous_tone(rate, freq, ampl, desired_size=None, max_size=None):
    """
    Return a buffer containing a complex tone at frequency freq. The tone is
    continuous, that is, repeating this signal will produce a continuous phase
    sinusoid.
    The buffer will try and approximate desired_size in length. If it is not
    possible to create a buffer smaller than max_size, an exception is thrown.

    Arguments:
    rate   -- Sampling rate in Hz.
    freq   -- Tone frequency in Hz
    ampl   -- Amplitude
    desired_size -- Number of samples ideally in returned buffer
    max_size -- Number of samples maximally in returned buffer
    """
    desired_size = desired_size or 1.0 * rate # About one second worth of data
    max_size = max_size or 100e6
    assert rate > freq
    rate_int = int(rate)
    freq_int = int(freq)
    gcd = math.gcd(rate_int, freq_int)
    rate_int = rate_int / gcd
    freq_int = freq_int / gcd
    length = max(freq_int * rate_int, 1) # freq may be zero
    tone = \
        numpy.exp(1j * 2 * numpy.pi * (freq/rate) * numpy.arange(length), dtype=numpy.complex64) \
        * ampl
    if length < desired_size:
        tone = numpy.tile(tone, int(desired_size // length))
    if length > max_size:
        raise ValueError("Cannot create a TX buffer! Rate/Freq ratio is too odd.")
    return tone

def get_power_dbfs(signal):
    """
    Return the power in dBFS for a digital signal, where fullscale is considered
    1.0.

    A sinusoid with amplitude 1.0 will thus return 0.0 (dBFS).
    """
    return 10 * numpy.log10(numpy.var(signal))

def get_usrp_power(streamer, num_samps=1e6, chan=0):
    """
    Return the measured input power in dBFS

    The return value is a list of dBFS power values, one per channel.
    """
    recv_buffer = numpy.zeros(
        (streamer.get_num_channels(), num_samps), dtype=numpy.complex64)
    metadata = uhd.types.RXMetadata()
    stream_cmd = uhd.types.StreamCMD(uhd.types.StreamMode.num_done)
    stream_cmd.num_samps = num_samps
    stream_cmd.stream_now = True
    streamer.issue_stream_cmd(stream_cmd)
    # Pass in long timeout, so we can rx the entire buffer in one go
    samps_recvd = streamer.recv(recv_buffer, metadata, 5.0)
    if samps_recvd != num_samps:
        raise RuntimeError(
            "ERROR! get_usrp_power(): Did not receive the correct number of samples!")
    return get_power_dbfs(recv_buffer[chan])



#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Import cal types into Python
"""

from ... import libpyuhd as lib

# Disable PyLint because we want to make it look like the following classes are
# defined in Python, but they're just renames of lib types. They therefore
# follow name conventions for Python classes, not for global constants.
# pylint: disable=invalid-name
# database is a class, but we treat it like a namespace, i.e., a submodule
database = lib.cal.database
Source = lib.cal.source
IQCal = lib.cal.iq_cal
PwrCal = lib.cal.pwr_cal
InterpMode = lib.cal.interp_mode
# pylint: enable=invalid-name



"""
Copyright 2019 Ettus Research, A National Instrument Brand

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import logging
import os
import sys
from ruamel import yaml

# Allow jsonschema import to fail. If not available no schema validation will
# be done (but warning will be printed for each skipped validation).
try:
    import jsonschema
except ImportError:
    logging.warning("Module jsonschema is not installed. Configuration files "
                    "will not be validated against their schema.")


def find_schema(schema_name, config_path):
    """
    Recursive search for schema file. Only looks for a file with appropriate
    name without checking for content or accessibility.
    This check will be performed later when trying to load the schema.
    :param schema_name: name of schema file to search for
    :param config_path: root path to start search in
    :return: full path to schema file if found, None else
    """
    for root, _, _ in os.walk(config_path):
        filename = os.path.join(root, schema_name)
        if os.path.isfile(filename):
            return filename
    return None


def validate_config(config, config_path):
    """
    Try to validate config.

    config contains a configuration loaded from a yaml file. config is assumed
    to be a dictionary which contains a key 'schema' which determines
    which schema to validate against. The schema (json formatted) needs to be
    located in config_path or any sub folder of it.
    If "jsonschema" module cannot be loaded validation is skipped and config
    is assumed to be valid. This way a configuration can also be loaded if
    "jsonschema" is not available.
    The method raises ValueError if no schema is defined in config or the
    schema defined in config cannot be found. The validation itself may throw
    a jsonschema.exceptions.ValidationError if config does not confirm to its
    schema.
    :param config: a dictionary to validate (loaded from yaml file).
    :param config_path: a path holding schema definitions
    """
    if "jsonschema" not in sys.modules:
        logging.warning("Skip schema validation (missing module jsonschema).")
        return

    if not "schema" in config:
        raise ValueError("Missing schema in configuration.")

    schema_name = config["schema"]
    logging.debug("Validate against schema %s.", schema_name)

    schema_file = find_schema('%s.json' % schema_name, config_path)
    if not schema_file:
        raise ValueError("Unknown schema %s." % schema_name)

    logging.debug("Using schema file %s.", schema_file)

    with open(schema_file) as stream:
        jsonschema.validate(instance=config, schema=json.load(stream))
        logging.debug("Configuration successful validated.")


def load_config(config_file, config_path):
    """
    Wrapper method to unify loading of configuration files.
    Beside loading the configuration (yaml file format) itself from config_file
    this method also validates the configuration against a schema. The root
    element of the configuration must load to a dictionary which contains a
    "schema" key. The value of schema points to a file named {value}.json which
    must be located in config_path or one of its sub folders.
    .. seealso:: validate_config.
    :param config_file: configuration to load
    :param config_path: root path of schema definition files
    :return:
    """
    logging.debug("Load configuration %s.", config_file)
    with open(config_file) as stream:
        rt_yaml = yaml.YAML(typ='rt')
        config = rt_yaml.load(stream)
        logging.debug("Configuration successful loaded.")
        validate_config(config, config_path)
        return config
